# Import internal libraries
from datetime import date

# Import external libraries
import pandas as pd

# Import helper functions
from heat_helper.core import _parse_year_group_to_int, CURRENT_ACADEMIC_YEAR_START
from heat_helper.exceptions import InvalidYearGroupError, FELevelError


def reverse_date(input_date: date, errors: str = "raise") -> date:
    """Sometimes dates are incorrectly formatted by Excel such that the day and month is swapped around. This can create errors when reading the data into pandas DataFrames. This function can be used to create a 'reversed' date where the day and month are swapped around. If this creates a date which doesn't exist, the original date is returned.

    Args:
        input_date: The date you wish to 'reverse' (swap day and month).
        errors: Defaults to 'raise' which raises errors. 'ignore' ignores errors and returns original date.

    Raises:
        TypeError: Raised if input_date is not in the date format (or pandas datetime format.)

    Returns:
        date: Reversed date or original date if reversed date does not exist.
    """
    try:
        if pd.isna(input_date):
            return input_date
        if not isinstance(input_date, date):
            raise TypeError(
                f"input_date must be date format, not {type(input_date).__name__}"
            )
        if input_date.day > 12:
            return input_date
        else:
            return input_date.replace(day=input_date.month, month=input_date.day)
    except (TypeError, NameError):
        if errors == "ignore":
            return input_date
        raise


def calculate_dob_range_from_year_group(
    year_group: str | int | pd.Series,
    start_year: int = CURRENT_ACADEMIC_YEAR_START,
    errors: str = "raise",
) -> tuple[date | None, date | None] | tuple[pd.Series, pd.Series]:
    """Calculates the expected DOB range (Sep 1 to Aug 31) for a given year group (1 to 13) in England.
    Includes some logic to try to handle Reception if entered as 'Reception', 'R', or 'Year R'.

    Args:
        year_group: The year group you want to find the date of birth range for. Examples: 'Year 10', 'Y10', 10. Note: Reception should be entered as Reception, Year R or R.
        start_year (optional): The year in which the academic year starts for the academic year you want to calculate. Example: for 2025/2026 enter 2025. You can enter any year here and it will return the date of birth range for someone in that year group during the specified academic year. Default is start of current academic year.
        errors (optional): default = 'raise' which raises all errors. 'ignore' and 'coerce' returns None, None.

    Raises:
        InvalidYearGroupError: Raised when `year_group` input cannot be parsed or is out of range.
        SchoolYearError: Raised when `start_year` is not a valid int.
        FELevelError: Raised if FE Levels are in `year_group`.

    Returns:
        The date of birth range. First date is start of the academic year; second date is the end of the academic year. Example: 01/09/2013, 31/08/2014."""
    
    try:
        # Dataframe
        if isinstance(year_group, pd.Series):
            results = year_group.apply(
                calculate_dob_range_from_year_group, 
                start_year=start_year, 
                errors=errors
            )
            
            start_dates, end_dates = zip(*results)
            return (pd.Series(start_dates, index=year_group.index), 
                    pd.Series(end_dates, index=year_group.index))

        # Individual values
        y_num = _parse_year_group_to_int(year_group)
        dob_start_year = int(start_year) - (y_num + 5)
        return date(dob_start_year, 9, 1), date(dob_start_year + 1, 8, 31)
    except (InvalidYearGroupError, TypeError, ValueError, FELevelError):
        if errors == "coerce":
            return None, None
        if errors == "ignore":
            return None, None  # Dates usually can't be 'ignored' as strings
        raise



# Import utils

from .utils import (
    get_excel_filepaths_in_folder, 
    convert_col_snake_case
)

from .names import (
    format_name,
    find_numbers_in_text,
    remove_numbers,
    create_full_name,
    remove_diacritics,
    remove_punctuation,
)

from .dates import (
    reverse_date, 
    calculate_dob_range_from_year_group
)

from .postcode import (
    format_postcode,
)

from .yeargroup import (
    clean_year_group,
    calculate_year_group_from_date,
)

from .matching import (
    perform_exact_match,
    perform_fuzzy_match,
    perform_school_age_range_fuzzy_match,
)

from .updates import (
    get_updates,
    get_contextual_updates,
)

from .duplicates import (
    find_duplicates,
)

# Import *
__all__ = [
    "calculate_dob_range_from_year_group",
    "clean_year_group",
    "format_postcode",
    "get_excel_filepaths_in_folder",
    "format_name",
    "find_numbers_in_text",
    "remove_numbers",
    "reverse_date",
    "create_full_name",
    "remove_diacritics",
    "perform_exact_match",
    "calculate_year_group_from_date",
    "perform_fuzzy_match",
    "perform_school_age_range_fuzzy_match",
    "get_updates",
    "get_contextual_updates",
    "convert_col_snake_case",
    "find_duplicates",
    "remove_punctuation"
]

# Version constant
__version__ = "0.1.4"



# Import internal libraries
import re
from datetime import date
import pandas as pd

# Import external libraries

# Import helper functions
from heat_helper.exceptions import InvalidYearGroupError, FELevelError


# Get CURRENT_ACADEMIC_YEAR_START constant function defined here apart from others due to constant below
def _calc_current_academic_year_start(date_now: date) -> int:
    if date_now.month in [9, 10, 11, 12]:
        return date_now.year
    else:
        return date_now.year - 1


# CONSTANTS
# Used in clean year groups
RECEPTION_ALIASES = {"reception", "r", "year r", "rec", "year group r", "y0", "year 0"}

# Used to validated postcode format
POSTCODE_REGEX = r"^[A-Z]{1,2}[0-9][A-Z0-9]? [0-9][A-Z]{2}$"

# Used to calculate current academic year for year group / date maniuplation functions
CURRENT_ACADEMIC_YEAR_START = _calc_current_academic_year_start(date.today())

# Used to remove punctuation except hyphens and apostrophes
PUNCTUATION = '!@#£$%^&*()_=+`~,.<>/?;:"\\|[]'

# For matching functions
STUDENT_HEAT_ID = "Student HEAT ID"


# Helper functions for main functions
def _parse_year_group_to_int(year_group: str | int | pd.Series) -> int:
    """Internal helper to convert any year group input to an integer (0-13)."""
    if isinstance(year_group, pd.Series):
        return year_group.apply(_parse_year_group_to_int)
    if isinstance(year_group, str):
        if "level" in year_group.lower():
            raise FELevelError(year_group)
        clean_input = year_group.strip().lower()
        if clean_input in RECEPTION_ALIASES:
            return 0
        match = re.search(r"\d+", clean_input)
        if not match:
            raise InvalidYearGroupError(year_group)
        y_num = int(match.group())
    elif isinstance(year_group, int):
        y_num = year_group
    else:
        raise TypeError(f"Input must be str or int, not {type(year_group).__name__}")

    if not (0 <= y_num <= 13):
        raise InvalidYearGroupError(year_group)
    return y_num


def _string_contains_int(string: str) -> bool:
    match = re.search(r"[0-9]+", string)
    if not match:
        return False
    else:
        return True


def _is_valid_postcode(postcode: str) -> bool:
    """Checks if a string is a validly formatted UK postcode. Does not check a postcode exists.
    Matches formats: A9 9AA, A99 9AA, AA9 9AA, AA99 9AA, A9A 9AA, AA9A 9AA.

    Args:
        postcode: the postcode to pattern match.

    Returns:
        True/False
    """
    if not isinstance(postcode, str):
        return False

    # If it's too short to even be a postcode, fail fast
    if len(postcode) < 5:
        return False

    # Check against the Regex
    return bool(re.match(POSTCODE_REGEX, postcode))


def _to_snake(name: str) -> str:
    # 1. Handle CamelCase (e.g., FirstName -> First_Name)
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name.strip())
    s2 = re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1)

    # 2. Remove special characters (keep only alphanumeric and spaces)
    clean = re.sub(r"[^a-zA-Z0-9\s_]", "", s2)

    # 3. Collapse whitespace, lower, and underscore
    return re.sub(r"[_\s]+", "_", clean.strip()).lower()


# Internal python libraries
from datetime import date

# Import helper functions
from heat_helper.exceptions import InvalidYearGroupError
from heat_helper.core import _parse_year_group_to_int, CURRENT_ACADEMIC_YEAR_START


def clean_year_group(year_group: str | int, errors: str = "raise") -> str | None:
    """Takes school year groups and cleans them to have the consistent format 'Year i'.

    Args:
        year_group: Text you wish to clean. Numbers entered will be cast to strings if possible.
        errors: default = 'raise' which raises all errors. 'ignore' returns original value, 'coerce' returns None.

    Raises:
        InvalidYearGroupError: Raised when the year group input cannot be parsed or is out of range.
        SchoolYearError: Raised when year_group is not a valid string.

    Returns:
        Cleaned year group in the format 'Year i'.
    """
    try:
        y_num = _parse_year_group_to_int(year_group)
        return "Reception" if y_num == 0 else f"Year {y_num}"
    except (InvalidYearGroupError, TypeError, ValueError):
        if errors == "ignore":
            return str(year_group)
        if errors == "coerce":
            return None
        raise


def calculate_year_group_from_date(
    input_date: date,
    start_of_academic_year: int = CURRENT_ACADEMIC_YEAR_START,
    errors: str = "raise",
) -> str | None:
    """Calculates school year group from date of birth for the English school system. Returns 'Year i' or 'Reception', or 'Student too young for school' if date of birth is not of school age.

    Args:
        input_date: Date of birth you wish to know the school year for.
        start_of_academic_year (optional): The school year in which you want to calculate the year group for. Allows you to calculate a year group for any academic year not just current e.g. for 2025/26 school year enter 2025. Default is start of current academic year.

    Raises:
        TypeError: Raised if input_date is not a date.

    Returns:
        Returns 'Year i' or 'Reception' or 'Student too young for school'.
    """
    try:
        if not isinstance(input_date, date):
            raise TypeError(f"Input must be a date, not {type(input_date).__name__}")
        if input_date.month in [9, 10, 11, 12]:
            offset = 5
        else:
            offset = 4
        year_group = start_of_academic_year - input_date.year - offset
        if year_group == 0:
            return "Reception"
        if year_group < 0:
            return "Student too young for school"
        if year_group > 13:
            raise InvalidYearGroupError(f"Year {year_group}")
        else:
            return f"Year {year_group}"
    except:
        if errors == "ignore":
            return None
        raise



import pandas as pd
from rapidfuzz import fuzz, process

from heat_helper.exceptions import ColumnDoesNotExistError


def find_duplicates(
    df: pd.DataFrame,
    name_col: str | list[str],
    date_of_birth_col: str,
    postcode_col: str,
    id_col: str = None,
    threshold: int = 80,
    fuzzy_type: str = "permissive",
    twin_protection: bool = True,
) -> pd.DataFrame:
    """Attempts to find duplicate records within one DataFrame. 
    The function looks for exact matches on any columns passed to name_col, date_of_birth_col and postcode_col, 
    and then attempts to fuzzy match names using either date_of_birth_col or date_of_birth_col and postcode_col 
    to create blocks of potential matches. Strictness of duplicate matching can be controlled using threshold 
    (% match for fuzzy name matching), fuzzy type (permission or strict) which pools potential duplicates for matching by
    using either date of birth or date of birth and postcode, and setting twin_protection to True/False. Twin Protection isolates
    first names in potential matches to filter out people with totally different first names. This is not totally failsafe and may
    still return some twins as potential duplicates.

    Args:
        df (pd.DataFrame): The DataFrame contain records to check for duplicates.
        name_col (str | list[str]): The column or list of columns contain names. Pass a list in the order the columns should be joined to create a full name e.g. ['First Name', 'Middle Name', 'Last Name'].
        date_of_birth_col (str): The column containing date of birth.
        postcode_col (str): The column containing postcode.
        id_col (str, optional): If there is already a column in your DataFrame which contains some kind of ID number, set it here. Otherwise, one will be created. Defaults to None.
        threshold (int, optional): The threshold for fuzzy matching. The percentage match of the name. Defaults to 80.
        fuzzy_type (str, optional): Controls whether date_of_birth_col or date_of_birth_col and postcode_col are used to create blocks for fuzzy matching. 'permissive' uses only date_of_birth_col, so will find duplicates with different postcodes. 'strict' uses both columns, so will only return potential duplicates where both date of birth and postcode match. Defaults to "permissive".
        twin_protection (bool, optional): If True, this filters out suspected twins with less similar first names (<65% match) from returned potential duplicates. Defaults to True.

    Raises:
        TypeError: Raised if df is not a DataFrame.
        ValueError: Raised if threshold is not a value between 0 and 100 or if fuzzy_type is not 'strict' or 'permissive'.
        ColumnDoesNotExistError: Raised if any of the columns passed as args are not in df.

    Returns:
        A DataFrame with a column called 'Potential Duplicates' which contains a list of IDs for any potential duplicates found by the function.
    """
    # Error Handling
    if not isinstance(df, pd.DataFrame):
        raise TypeError(f"{df} is not a DataFrame")

    if not (0 <= threshold <= 100):
        raise ValueError("Threshold must be an integer between 0 and 100")

    if fuzzy_type not in ["strict", "permissive"]:
        raise ValueError("fuzzy_type must be 'strict' or 'permissive'")
    
    # Check cols exist
    if isinstance(name_col, str):
        if name_col not in df.columns:
            raise ColumnDoesNotExistError(f"'{name_col}' not found in {df} columns")
    else:
        for name in name_col:
            if name not in df.columns:
                raise ColumnDoesNotExistError(f"'{name_col}' not found in {df} columns")
    
    if date_of_birth_col not in df.columns:
            raise ColumnDoesNotExistError(f"'{date_of_birth_col}' not found in {df} columns")
    
    if postcode_col not in df.columns:
            raise ColumnDoesNotExistError(f"'{postcode_col}' not found in {df} columns")

    new_df = df.copy()

    # Set up col list depending on if name_col is a list or single str
    if isinstance(name_col, list):
        new_df["_match_name"] = (
            new_df[name_col].fillna("").astype(str).agg(" ".join, axis=1)
        )
        col_list = name_col + [date_of_birth_col, postcode_col]
    else:
        col_list = [name_col, date_of_birth_col, postcode_col]
        new_df["_match_name"] = new_df[name_col]

    # String Column Cleaning
    for col in col_list:
        if new_df[col].dtype == "object":
            new_df[col] = (
                df[col].str.strip().replace(r"\s+", " ", regex=True).fillna("")
            )

    # Set up ID column if not passed to function
    if id_col is None:
        new_df["Duplicate ID"] = "#" + (pd.Series(range(len(new_df))) + 1).astype(str)
        id_col = "Duplicate ID"

    print("Searching for duplicates...")

    # 1. Exact Matches
    def _format_duplicate_list(group_series):
        # If there is more than one item in the group, it's a duplicate
        if len(group_series) > 1:
            return ", ".join(sorted(group_series))
        return ""
    
    new_df["Potential Duplicates"] = (
        new_df.groupby(col_list)[id_col].transform(_format_duplicate_list).fillna("")
    )

    # 2. Fuzzy Matching
    # We use a custom Union-Find structure to track clusters.
    # parent[x] = y means "x belongs to the same cluster as y"
    parent = {}

    def find_root(i):
        # Recursively find the root representative of ID i
        if parent.setdefault(i, i) != i:
            parent[i] = find_root(parent[i])  # Path compression
        return parent[i]

    def union(i, j):
        # Merge the sets containing i and j
        root_i = find_root(i)
        root_j = find_root(j)
        if root_i != root_j:
            parent[root_i] = root_j

    # A. Register Exact Matches into Union-Find
    # If we already found #1 and #2 are exact matches, union them now.
    has_dupe = new_df[new_df["Potential Duplicates"] != ""]
    for row_ids in has_dupe["Potential Duplicates"]:
        ids_in_group = row_ids.split(", ")
        first_id = ids_in_group[0]
        for other_id in ids_in_group[1:]:
            union(first_id, other_id)

    # B. Run Fuzzy Matching
    if fuzzy_type == "strict":
        blocks = new_df.groupby([date_of_birth_col, postcode_col])
    else:
        blocks = new_df.groupby(date_of_birth_col)

    for _, block_df in blocks:
        if len(block_df) < 2:
            continue

        names = block_df["_match_name"].tolist()
        ids = block_df[id_col].tolist()

        # Calculate similarity matrix
        score_matrix = process.cdist(names, names, scorer=fuzz.token_sort_ratio)

        # Iterate upper triangle to avoid duplicate checks
        for i in range(len(score_matrix)):
            for j in range(i + 1, len(score_matrix)):
                if score_matrix[i][j] >= threshold:
                    if twin_protection:
                        # Split name on first space. 0 = Name before first space
                        name_1 = names[i].split(" ")[0]
                        name_2 = names[j].split(" ")[0]

                        # Compare ONLY the first names
                        first_name_score = fuzz.ratio(name_1, name_2)

                        # If the full strings match, but the first names are clearly different,
                        # assume they are twins (or siblings) and SKIP the union.
                        if (
                            first_name_score < 70
                        ):
                            continue

                    # If fuzzy match found, Union the two IDs
                    union(ids[i], ids[j])

    # --- 5. Final Reconciliation ---
    # Now we simply group all IDs by their "Root Parent"
    # This automatically handles the A->B->C chaining
    clusters = {}
    for i in new_df[id_col]:
        root = find_root(i)
        if root not in clusters:
            clusters[root] = []
        clusters[root].append(i)

    # Convert clusters to string format: "#1, #2, #3"
    # We only care about clusters with size > 1
    id_to_string_map = {}
    for root, members in clusters.items():
        if len(members) > 1:
            member_str = ", ".join(sorted(members))
            for member in members:
                id_to_string_map[member] = member_str

    # Apply the map
    new_df['Potential Duplicates'] = new_df[id_col].map(id_to_string_map).fillna("")

    # Final clean up
    if "_match_name" in new_df.columns:
        new_df.drop(columns=["_match_name"], inplace=True)

    new_df['Potential Duplicates'] = new_df['Potential Duplicates'].replace(
        r"^\s*$", None, regex=True
    )

    new_df = new_df.sort_values(['Potential Duplicates', id_col], ascending=False)

    dupe_count = len(new_df) - new_df['Potential Duplicates'].isna().sum()
    print(f"{dupe_count} records are potential duplicates.")

    return new_df



import pandas as pd
from rapidfuzz import process, fuzz

from heat_helper.exceptions import (
    ColumnDoesNotExistError,
    FuzzyMatchIndexError,
    FilterColumnMismatchError,
)
from heat_helper.dates import calculate_dob_range_from_year_group
from heat_helper.core import CURRENT_ACADEMIC_YEAR_START, STUDENT_HEAT_ID


def perform_exact_match(
    unmatched_df: pd.DataFrame,
    heat_df: pd.DataFrame,
    left_join_cols: list[str],
    right_join_cols: list[str],
    match_desc: str,
    verify: bool = False,
    heat_id_col: str = STUDENT_HEAT_ID,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Performs an exact match on specified columns between new data and your HEAT Student export and returns the HEAT Student ID if a match is found.
    This function returns two DataFrames: one containing the matches and one containing unmatched students, for passing to another matching function.
    This is useful to create a matching waterfall where you move through different levels of strictness.


    Args:
        unmatched_df: The DataFrame containing the students you want to search for.
        heat_df: The DataFrame containing your HEAT Student Export.
        left_join_cols: Columns in new_df you want to match on.
        right_join_cols: Columns in heat_df you want to match on.
        match_desc: A description of the match; added to a 'Match Type' col in the returned matched DataFrame. Should be descriptive to help you verify matches later, especially if joining multiple returns of this function and exporting to a .csv or Excel file.
        verify (optional): Defaults to False. Controls whether to return all columns from heat_df to the matched DataFrame for verifying of matches. Useful if you are performing a less exact match and you want to verify the returned students. Also useful if you are using this function or perform_fuzzy_match function and want to join results together (column structure will be the same).
        heat_id_col (optional): Defaults to 'Student HEAT ID'. Use this if the column in your HEAT Export with the Student ID in is not called 'Student HEAT ID'.

    Raises:
        TypeError: Raised if new_df or heat_df are not pandas DataFrames.
        ColumnDoesNotExistError: Raised if a column you are trying to use for matching does not exist in either new_df or heat_df.

    Returns:
        Two DataFrames: first DataFrame is matched data, second is remaining data for onward matching.
    """

    # Checks before function starts
    if not isinstance(unmatched_df, pd.DataFrame) or not isinstance(
        heat_df, pd.DataFrame
    ):
        raise TypeError("new_df and heat_df must be pandas DataFrames.")
    # Check cols exist
    for col in left_join_cols:
        if col not in unmatched_df.columns:
            raise ColumnDoesNotExistError(f"'{col}' not found in new_df")
    for col in right_join_cols:
        if col not in heat_df.columns:
            raise ColumnDoesNotExistError(f"'{col}' not found in heat_df")
    if heat_id_col not in heat_df.columns:
        raise ColumnDoesNotExistError(
            f"Specified ID column '{heat_id_col}' not found in heat_df."
        )

    if unmatched_df.empty:
        print(
            f"WARNING: skipping match type: {match_desc} - no students left to match."
        )
        return pd.DataFrame(), unmatched_df
    else:
        # For performance, heat_df should just be columns required for match + heat ID
        heat_cols = list(right_join_cols)
        heat_cols_list = heat_cols + [heat_id_col]
        heat_df_slim = heat_df[heat_cols_list]

        # Initial slim merge using only data req. for match
        joined_df = pd.merge(
            unmatched_df,
            heat_df_slim,
            left_on=left_join_cols,
            right_on=right_join_cols,
            how="left",
            suffixes=("", "_match"),
        )

        # Separate matches and non-matches
        final_matched = (
            joined_df.dropna(subset=[heat_id_col]).copy().reset_index(drop=True)
        )
        final_matched["Match Type"] = match_desc

        unmatched = (
            joined_df[joined_df[heat_id_col].isnull()]
            .copy()
            .reset_index(drop=True)
        )
        unmatched = unmatched[unmatched_df.columns]

        # Reporting to terminal
        total_new = len(unmatched_df)
        total_unmatched = len(unmatched)
        students_matched_count = total_new - total_unmatched
        has_duplicates = len(final_matched) > students_matched_count
        print(f"   Attempting to match {total_new} students. Match type: {match_desc}.")
        print(f"     ...{students_matched_count} students found in HEAT data")
        print(f"     ...{len(unmatched)} students left to find.")
        if has_duplicates:
            diff = len(final_matched) - students_matched_count
            print(
                f"     WARNING: {diff} extra record(s) created. Some student matched to multiple HEAT records. Check HEAT data for duplicates."
            )

        if verify:
            rename_dict = {
                col: f"HEAT: {col}"
                for col in heat_df.columns
                if col != heat_id_col
            }
            heat_df_verif = heat_df.rename(columns=rename_dict)
            final_matched_check = pd.merge(
                final_matched, heat_df_verif, how="left", on=heat_id_col
            )
            final_matched_check = final_matched_check.rename(
                columns={heat_id_col: f"HEAT: {heat_id_col}"}
            )
            return final_matched_check, unmatched
        else:
            cols_list = unmatched_df.columns
            cols_list = list(cols_list)
            cols_list.extend(["Match Type", heat_id_col])
            final_matched = final_matched[cols_list]
            final_matched = final_matched.rename(
                columns={heat_id_col: f"HEAT: {heat_id_col}"}
            )
            return final_matched, unmatched


def perform_fuzzy_match(
    unmatched_df: pd.DataFrame,
    heat_df: pd.DataFrame,
    left_filter_cols: list[str],
    right_filter_cols: list[str],
    left_name_col: str,
    right_name_col: str,
    match_desc: str,
    threshold: int = 80,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """This function allows you to fuzzy match names of students in an external dataset to your HEAT Student Export to retrieve HEAT Student IDs.
    You can control the potential pool of fuzzy matches by specifying filter columns in both DataFrames e.g. only look for fuzzy matches where Date of Birth and Postcode matches.

    Args:
        unmatched_df (pd.DataFrame): The DataFrame of students you want to fuzzy match.
        heat_df (pd.DataFrame): The DataFrame containing your HEAT Student Export.
        left_filter_cols: Filter columns in unmatched_df. By specifying a column here it will be used to control the pool of possible fuzzy matches. For example, by setting Date of birth and postcode here, it will only fuzzy match 'Jo Smith' to 'Joanne Smith' if both records have the same date of birth and postcode.
        right_filter_cols: Corresponding filter columns in heat_df. Must match those set in left_filter_cols.
        left_name_col: Column which contains the name information (to be matched) in unmatched_df.
        right_name_col: Column which contains the name information in heat_df.
        match_desc: A description of the match; added to a 'Match Type' col in the returned matched DataFrame. Should be descriptive to help you verify matches later, especially if joining multiple returns of this function and exporting to a .csv or Excel file.
        threshold (optional): The acceptable percentage match for fuzzy matching. Higher is stricter and matches will be more similar. Defaults to 80.

    Raises:
        TypeError: Raised if unmatched_df or heat_df are not pandas DataFrames.
        ColumnDoesNotExistError: Raised if columns specified as filters or name columns do not exist in their DataFrames.
        FilterColumnMismatchError: Raised if unequal number of columns specified in left and right filters.
        FuzzyMatchIndexError: Raised when unmatched_df does not have a unique index and cannot be used for matching.

    Returns:
        Two DataFrames: first DataFrame is matched data, second is remaining data for onward matching.
    """
    # Type checking and error handling:
    if not isinstance(unmatched_df, pd.DataFrame) or not isinstance(
        heat_df, pd.DataFrame
    ):
        raise TypeError("unmatched_df and heat_df must be pandas DataFrames.")
    # Check cols exist
    for col in left_filter_cols:
        if col not in unmatched_df.columns:
            raise ColumnDoesNotExistError(f"'{col}' not found in unmatched_df")
    for col in right_filter_cols:
        if col not in heat_df.columns:
            raise ColumnDoesNotExistError(f"'{col}' not found in heat_df")
    if left_name_col not in unmatched_df.columns:
        raise ColumnDoesNotExistError(f"'{left_name_col}' not found in unmatched_df")
    if right_name_col not in heat_df.columns:
        raise ColumnDoesNotExistError(f"'{right_name_col}' not found in heat_df")
    # Check filter cols are same length
    if len(left_filter_cols) != len(right_filter_cols):
        raise FilterColumnMismatchError(
            "left_filter_cols and right_filter_cols must have the same length for mapping."
        )
    # Check unmatched has a unique index
    if not unmatched_df.index.is_unique:
        raise FuzzyMatchIndexError("unmatched_df")

    # Warning about column collisions
    collision_cols = [c for c in unmatched_df.columns if c.endswith("_HEAT")]

    if collision_cols:
        print(
            f"WARNING: The input unmatched_df contains columns that already end in '_HEAT': {collision_cols}. These will be renamed to 'HEAT: ...' in the final output and may be indistinguishable from actual data retrieved from the HEAT database."
        )

    if unmatched_df.empty:
        print(
            f"WARNING: skipping match type: {match_desc} - no students left to match."
        )
        return pd.DataFrame(), unmatched_df
    else:
        print(
            f"     Attempting to match {len(unmatched_df)} students. Fuzzy match type: {match_desc}."
        )

        # Create copies in case slice passed to function
        unmatched_df = unmatched_df.copy()
        heat_df = heat_df.copy()

        # create heat_df blocks for faster matching
        grouped_heat = heat_df.groupby(right_filter_cols).groups

        matched_results = []

        # Go through each row in unmatched data
        for idx, row in unmatched_df.iterrows():
            search_values = row[left_filter_cols].tolist()
            search_key = (
                tuple(search_values) if len(search_values) > 1 else search_values[0]
            )

            # Does this block exist in the Heat data?
            if search_key in grouped_heat:
                # Get the indices of the rows in the Heat data that match this block
                potential_match_indices = grouped_heat[search_key]
                potential_matches = heat_df.loc[potential_match_indices]

                # Fuzzy Match only within this specific block
                choices = potential_matches[right_name_col].to_dict()  # {index: name}

                best_match = process.extractOne(
                    query=row[left_name_col],
                    choices=choices,
                    scorer=fuzz.token_sort_ratio,
                    score_cutoff=threshold,
                )

                if best_match:
                    name, score, heat_idx = best_match
                    # Reconstruct the row
                    res = pd.concat([row, heat_df.loc[heat_idx].add_suffix("_HEAT")])
                    res["Fuzzy Score"] = round(score, 2)
                    res["Match Type"] = match_desc
                    res["__SOURCE_INDEX__"] = idx
                    matched_results.append(res)

        # final_matches processing
        final_matches = pd.DataFrame(matched_results)
        if not final_matches.empty:
            final_matches.sort_values(
                by="Fuzzy Score", ascending=False, inplace=True, ignore_index=True
            )

            # Rename HEAT columns
            heat_cols = [c for c in final_matches.columns if c.endswith("_HEAT")]
            mapping = {col: f"HEAT: {col.removesuffix('_HEAT')}" for col in heat_cols}
            final_matches.rename(columns=mapping, inplace=True)

            # Sort out indices for dropping
            matched_indices = final_matches["__SOURCE_INDEX__"].tolist()
            final_matches.drop(columns=["__SOURCE_INDEX__"], inplace=True)

            print(f"     ...{len(final_matches)} students found in HEAT data.")
        else:
            matched_indices = []
            print("     ...0 students found in HEAT data.")

        # Identify who is still missing
        remaining_unmatched = unmatched_df.drop(matched_indices)
        print(f"     ...{len(remaining_unmatched)} students left to find.")
        return final_matches, remaining_unmatched


def perform_school_age_range_fuzzy_match(
    unmatched_df: pd.DataFrame,
    heat_df: pd.DataFrame,
    unmatched_school_col: str,
    heat_school_col: str,
    unmatched_name_col: str,
    heat_name_col: str,
    unmatched_year_group_col: str,
    heat_dob_col: str,
    match_desc: str,
    heat_id_col: str = STUDENT_HEAT_ID,
    academic_year_start: int = CURRENT_ACADEMIC_YEAR_START,
    threshold: int = 80,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """This function attempts to fuzzy match the names of students to your HEAT data.
    To control the pool of fuzzy matches, data is first matched on school name, and then uses year group to only return students with a date of birth in range for that year group.
    Useful if you do not know a student's date of birth, but you do know which school they attend and their year group.
    Returns one dataframe of matches and one dataframe of remaining unmatched data.

    Args:
        unmatched_df: DataFrame containing student records you wish to fuzzy match to HEAT records.
        heat_df: DataFrame containing HEAT Student Export.
        unmatched_school_col: Column which contains School name in unmatched_df.
        heat_school_col: Column which contains school name in heat_df.
        unmatched_name_col: Column which contains Student name in unmatched_df.
        heat_name_col: Column which contains Student name in heat_df.
        unmatched_year_group_col: Column in unmatched_df which contains year group for age range calculation.
        heat_dob_col: Column in heat_df which contains Student Date of Birth.
        match_desc: A description of the match; added to a 'Match Type' col in the returned matched DataFrame. Should be descriptive to help you verify matches later, especially if joining multiple returns of this function and exporting to a .csv or Excel file.
        heat_id_col (optional): Column in heat_df which contains HEAT Student ID. Defaults to 'Student HEAT ID'.
        academic_year_start (optional): . Defaults to start of current academic year (calculated by package).
        threshold (optional): The acceptable percentage match for fuzzy matching. Higher is stricter and matches will be more similar. Defaults to 80.

    Raises:
        TypeError: Raised if unmatched_df or heat_df are not pandas DataFrames or if heat_dob_col is not in pandas Datetime format (will try to convert first.)
        ColumnDoesNotExistError: Raised if any specified column does not exist in its dataframe.
        FuzzyMatchIndexError: Raised if unmatched_df does not have unique index.

    Returns:
        Two DataFrames: first DataFrame is matched data, second is remaining data for onward matching.
    """
    # Type checking and error handling:
    if not isinstance(unmatched_df, pd.DataFrame) or not isinstance(
        heat_df, pd.DataFrame
    ):
        raise TypeError("unmatched_df and heat_df must be pandas DataFrames.")
    if not pd.api.types.is_datetime64_any_dtype(heat_df[heat_dob_col]):
        try:
            heat_df[heat_dob_col] = pd.to_datetime(heat_df[heat_dob_col]).dt.normalize()
            print(f"Note: Converted '{heat_dob_col}' to datetime format automatically.")
        except Exception:
            raise TypeError(
                f"'{heat_dob_col}' is not datetime and could not be converted."
            )
    # Check cols exist
    for col in [unmatched_school_col, unmatched_name_col, unmatched_year_group_col]:
        if col not in unmatched_df.columns:
            raise ColumnDoesNotExistError(f"'{col}' not found in unmatched_df")
    for col in [heat_name_col, heat_school_col, heat_dob_col]:
        if col not in heat_df.columns:
            raise ColumnDoesNotExistError(f"'{col}' not found in heat_df")

    # Check unmatched has a unique index
    if not unmatched_df.index.is_unique:
        raise FuzzyMatchIndexError("unmatched_df")

    # Copy originals in case df slice passed to this function; tidy up school names to improve matching
    unmatched_df = unmatched_df.copy()
    heat_df = heat_df.copy()

    heat_df[heat_school_col] = (
        heat_df[heat_school_col]
        .astype(str)
        .str.title()
        .str.strip()
        .replace(r"\s+", " ", regex=True)
    )
    unmatched_df[unmatched_school_col] = (
        unmatched_df[unmatched_school_col]
        .astype(str)
        .str.title()
        .str.strip()
        .replace(r"\s+", " ", regex=True)
    )

    # Create blocks on school in heat_df for quicker matching
    grouped_heat = heat_df.groupby(heat_school_col).groups

    matched_results = []

    # Go through each unmatched row
    for idx, row in unmatched_df.iterrows():
        school_key = row[unmatched_school_col]
        year_group = row[unmatched_year_group_col]

        # Calculate the DOB range for this specific student's year group
        try:
            dob_range = calculate_dob_range_from_year_group(
                year_group, academic_year_start
            )
        except Exception:
            continue

        if not dob_range or school_key not in grouped_heat:
            continue

        start_date, end_date = dob_range

        # Filter the School Block by Age
        # Get only the HEAT records for this school
        potential_matches = heat_df.loc[grouped_heat[school_key]]

        # Further narrow down by Date of Birth (Age Match)
        age_mask = (potential_matches[heat_dob_col].dt.date >= start_date) & (
            potential_matches[heat_dob_col].dt.date <= end_date
        )
        final_potentials = potential_matches[age_mask]

        if not final_potentials.empty:
            # Fuzzy Match names within the filtered school/age block
            choices = final_potentials[heat_name_col].to_dict()

            best_match = process.extractOne(
                query=row[unmatched_name_col],
                choices=choices,
                scorer=fuzz.token_sort_ratio,
                score_cutoff=threshold,
            )

            if best_match:
                name, score, heat_idx = best_match
                # Reconstruct row
                res = pd.concat([row, heat_df.loc[heat_idx].add_suffix("_HEAT")])
                res["Fuzzy Score"] = round(score, 2)
                res["Match Type"] = match_desc
                res["__SOURCE_INDEX__"] = idx
                matched_results.append(res)

    # Sorting, renaming and tidying
    final_matches = pd.DataFrame(matched_results)

    if not final_matches.empty:
        final_matches.sort_values(
            by="Fuzzy Score", ascending=False, inplace=True, ignore_index=True
        )

        heat_student_id_col = f"{heat_id_col}_HEAT"

        initial_count = len(final_matches)
        final_matches = final_matches.drop_duplicates(
            subset=[heat_student_id_col], keep="first"
        )
        conflicts_removed = initial_count - len(final_matches)

        if conflicts_removed > 0:
            print(
                f"     ...Removed {conflicts_removed} duplicate HEAT ID assignments (kept highest scores)."
            )

        # Rename HEAT columns
        heat_cols = [c for c in final_matches.columns if c.endswith("_HEAT")]
        mapping = {col: f"HEAT: {col.removesuffix('_HEAT')}" for col in heat_cols}
        final_matches.rename(columns=mapping, inplace=True)

        # Sort out indices for dropping
        matched_indices = final_matches["__SOURCE_INDEX__"].tolist()
        final_matches.drop(columns=["__SOURCE_INDEX__"], inplace=True)

        print(f"     ...{len(final_matches)} school/age fuzzy matches found.")
    else:
        matched_indices = []
        print("     ...0 matches found.")

    remaining_unmatched = unmatched_df.drop(matched_indices)

    return final_matches, remaining_unmatched



# Import internal libraries
from typing import Any


## Custom Errors
class HeatHelperError(Exception):
    """Base class for all exceptions in this package."""

    pass


class InvalidYearGroupError(HeatHelperError):
    """Raised when the year group input cannot be parsed or is out of range."""

    def __init__(self, value: Any):
        self.value = value
        self.message = f"Invalid year group: '{value}'. Must be R/Reception or 0-13."
        super().__init__(self.message)


class FELevelError(HeatHelperError):
    """Raised when the year group input cannot be parsed or is out of range."""

    def __init__(self, value: Any):
        self.value = value
        self.message = f"Invalid year group: '{value}'. Cannot translate FE Levels to school year groups."
        super().__init__(self.message)


class InvalidPostcodeError(HeatHelperError):
    """Raised when the postcode is not a valid format."""

    def __init__(self, value: Any):
        self.value = value
        self.message = f"Invalid postcode format: '{value}'"
        super().__init__(self.message)


class ColumnDoesNotExistError(HeatHelperError):
    """Raised when a column is not found in a dataframe."""

    def __init__(self, value: Any):
        self.value = value
        self.message = f"Column does not exist: {value}"
        super().__init__(self.message)


class FuzzyMatchIndexError(HeatHelperError):
    """Raised when dataframe used for fuzzy matching does not have a unique index (which would compromise returned results.)."""

    def __init__(self, value: Any):
        self.value = value
        self.message = f"Index of {value} contains duplicate entries and cannot be used for fuzzy matching."
        super().__init__(self.message)


class FilterColumnMismatchError(HeatHelperError):
    """Raised when dataframe used for fuzzy matching does not have a unique index (which would compromise returned results.)."""

    def __init__(self, value: Any):
        self.value = value
        self.message = f"Filter columns do not match: {value}"
        super().__init__(self.message)


class ColumnsNotUnique(HeatHelperError):
    """Raised when dataframe used for fuzzy matching does not have a unique index (which would compromise returned results.)."""

    def __init__(self, value: Any):
        self.value = value
        self.message = f"Column names are not unique: {value}"
        super().__init__(self.message)



import pandas as pd
from heat_helper.exceptions import ColumnDoesNotExistError


def get_updates(df: pd.DataFrame, new_col: str, heat_col: str) -> pd.Series:
    """Compares two DataFrame columns and returns a value if new_col is different to heat_col.
    This can help you identify where new data is different to existing HEAT records and therefore needs updating.
    It returns a new column which only contains data that needs to be updated - so it can be copied to the HEAT import template.

    Args:
        df (pd.DataFrame): DataFrame where the columns are located.
        new_col (str): The column which contains 'new' data. This would be the data you want to update on HEAT if it differs from values in heat_col.
        heat_col (str): The corresponding column in your HEAT export e.g. if you are checking if any postcodes need updating, both new_col and heat_col should contain postcodes.

    Raises:
        TypeError: Raised if df is not a DataFrame or new_col and heat_col are not strings (text).
        ColumnDoesNotExistError: Raised if either new_col or heat_col are not in df columns.

    Returns:
        A DataFrame column where rows contain the value from new_col if this is different to heat_col.
    """
    # Type checking
    if not isinstance(df, pd.DataFrame):
        raise TypeError(f"{df} must be a DataFrame.")
    if not isinstance(new_col, str):
        raise TypeError(f"Text must be a string, not {type(new_col).__name__}")
    if not isinstance(heat_col, str):
        raise TypeError(f"Text must be a string, not {type(heat_col).__name__}")

    for col in [new_col, heat_col]:
        if col not in df.columns:
            raise ColumnDoesNotExistError(f"'{col}' not found in DataFrame column list.")

    left = df[new_col]
    right = df[heat_col]

    # Warning if dtypes are different
    if left.dtype != right.dtype:
        print(
            f"WARNING: Type Mismatch: {new_col} is {left.dtype}, but {heat_col} is {right.dtype}"
        )

    # Standard equality check that handles nulls
    condition = ~(left.eq(right) | (left.isna() & right.isna()))

    # Fill blank strings with None, then return updated data if True, otherwise return null
    if left.dtype == 'object':
        df[new_col] = df[new_col].fillna('').replace(r"^\s*$", None, regex=True)
    update_col = df[new_col].where(condition, other=None)  # type: ignore

    return update_col


def get_contextual_updates(
    df: pd.DataFrame, new_col: str, heat_col: str, bad_values: list[str]
) -> pd.Series:
    """This function is similar to get_updates, except you can also pass a list of 'bad' values you do not want to override HEAT data.
    This can be useful if you want to ensure that data like 'Not available' or 'Unknown' does not overwrite previously collected 'good' values in the contextual data columns.

    Args:
        df (pd.DataFrame): DataFrame where the columns are located.
        new_col (str): The column which contains 'new' data. This would be the data you want to update on HEAT if it differs from values in heat_col.
        heat_col (str): The corresponding column in your HEAT export e.g. if you are checking if any postcodes need updating, both new_col and heat_col should contain postcodes.
        bad_values (list[str]): A list of values which should not overwrite 'good' data in your HEAT records. For example, if your new data contains 'Not available' or 'Unknown' but your HEAT records had values in these columns, you could pass ['Not available', 'Unknown'] to this variable, and these will not overwrite 'good' values in the heat_col.

    Raises:
        TypeError: Raised if df is not a DataFrame, new_col and heat_col are not strings (text), or bad_values is not a list.
        ColumnDoesNotExistError: Raised if either new_col or heat_col are not in df columns.

    Returns:
        A DataFrame column where rows contain the value from new_col if this is different to heat_col.
    """
    # Type checking
    if not isinstance(df, pd.DataFrame):
        raise TypeError(f"{df} must be a DataFrame.")
    if not isinstance(new_col, str):
        raise TypeError(f"Text must be a string, not {type(new_col).__name__}")
    if not isinstance(heat_col, str):
        raise TypeError(f"Text must be a string, not {type(heat_col).__name__}")
    if not isinstance(bad_values, list):
        raise TypeError(f"{bad_values} must be a List.")

    for col in [new_col, heat_col]:
        if col not in df.columns:
            raise ColumnDoesNotExistError(f"'{col}' not found in {df} column list.")

    left = df[new_col]
    right = df[heat_col]

    # Warning if dtypes are different
    if left.dtype != right.dtype:
        print(
            f"WARNING: Type Mismatch: {new_col} is {left.dtype}, but {heat_col} is {right.dtype}"
        )

    # Define Update conditions: if values are different and left is not a bad_value
    left_good = ~left.isin(bad_values) & left.notna()
    is_different = ~(left.eq(right) | (left.isna() & right.isna()))
    condition = is_different & left_good

    if left.dtype == 'object':
        df[new_col] = df[new_col].fillna('').replace(r"^\s*$", None, regex=True)
    update_col = df[new_col].where(condition, other=None)  # type: ignore

    return update_col



# Import internal libraries
import re
import unicodedata

# Import external libraries
import pandas as pd

# Import helper functions
from .core import _string_contains_int, PUNCTUATION


def format_name(text: str, errors: str = "raise") -> str | None:
    """Cleans the formatting of names. Strips extra whitespaces, converts to title case (with exceptions for names like McDonald and O'Reilly) and tidies any spaces around hyphens.

    Args:
        text: The name you wish to clean.
        errors (optional): Default = 'raise' which raises all errors. 'ignore' ignores errors and returns original value, 'coerce' returns None.

    Raises:
        TypeError: Raised if text is not a string.

    Returns:
        Cleaned text.
    """
    replacements = (
        (r"\s*-\s*", "-"),  # Cleans spaces around hyphens
        (r"\s+", " "),  # Cleans any number of spaces -> one space
    )
    try:
        if not isinstance(text, str):
            raise TypeError(f"Text must be a string, not {type(text).__name__}")
        working_text = text.strip().title()
        for pattern, replacement in replacements:
            working_text = re.sub(pattern, replacement, working_text)
        # Preserves capitalisation after Mc and O' names
        working_text = re.sub(
            r"\b(Mc)([a-z])", lambda m: m.group(1) + m.group(2).upper(), working_text
        )
        working_text = re.sub(
            r"\b(O')([a-z])", lambda m: m.group(1) + m.group(2).upper(), working_text
        )
        return working_text
    except TypeError:
        if errors == "ignore":
            return text
        if errors == "coerce":
            return None
        raise


def find_numbers_in_text(
    text: str, errors: str = "raise", convert_to_string: bool = False
) -> bool | str | None:
    """Checks if one or more numbers are present in a string. Numbers do not have to be consecutive.

    Args:
        text: The text to check for numbers.
        errors (optional): Default = 'raise' which raises all errors. 'ignore' ignores errors and returns original value, 'coerce' returns None.
        convert_to_string (optional): Tells the function to convert text datatype to string, if possible. Defaults to False.

    Raises:
        TypeError: Raised if text datatype is not string.

    Returns:
        True if string contains one or more numbers (0-9) or False if no numbers present.
    """
    try:
        if convert_to_string:
            text = str(text)
        if not isinstance(text, str):
            raise TypeError(f"Text must be a string, not {type(text).__name__}")
        check = _string_contains_int(text)
        return check
    except TypeError:
        if errors == "ignore":
            return text
        if errors == "coerce":
            return None
        raise


def remove_numbers(
    text: str, errors: str = "raise", convert_to_string: bool = False
) -> str | None:
    """Removes one or more numbers from a string (text). Numbers do not have to be consecutive.

    Args:
        text: The string you want to remove numbers from e.g. 'Jane Doe 43'
        errors (optional): Default = 'raise' which raises all errors. 'ignore' ignores errors and returns original value, 'coerce' returns None.
        convert_to_string (optional): Tells the function to convert text datatype to string, if possible. Defaults to False.

    Raises:
        TypeError: Raised if text is not a string.

    Returns:
        Text with numbers removed.
    """
    try:
        if convert_to_string:
            text = str(text)
        if not isinstance(text, str):
            raise TypeError(f"Text must be a string, not {type(text).__name__}")
        if _string_contains_int(text):
            clean = re.sub(r"[0-9]+", "", text)
            return clean.strip()
        else:
            return text
    except TypeError:
        if errors == "ignore":
            return text
        if errors == "coerce":
            return None
        raise


def create_full_name(
    first_name: str | pd.Series,
    last_name: str | pd.Series,
    middle_name: str | pd.Series = "",
) -> str | pd.Series:
    """Joins strings or pandas DataFrame columns into a 'Full Name' string or column of strings. Useful if you are going to be fuzzy matching names.

    Args:
        first_name: First name.
        last_name: Last name.
        middle_name (optional): Middle name. Defaults to a blank string or blank pd.Series.

    Returns:
        One string or Series of strings with all names joined.
    """
    if isinstance(first_name, pd.Series):
        if isinstance(middle_name, str) and middle_name == "":
            middle_name = pd.Series("", index=first_name.index)
        if isinstance(middle_name, pd.Series):
            middle_name = middle_name.fillna("")
        full_name_pd = first_name + " " + middle_name + " " + last_name
        full_name_pd = full_name_pd.str.replace(r"\s+", " ", regex=True).str.strip()
        return full_name_pd
    if isinstance(first_name, str):
        name_parts = [
            str(part).strip() for part in [first_name, middle_name, last_name]
        ]
        full_name = " ".join(name_parts).strip()
        full_name = re.sub(r"\s+", " ", full_name)
        return full_name


def remove_diacritics(input_text: str, errors: str = "raise") -> str | None:
    """Removes diacritics (accented letters) from text. Uses python's built-in unicodedata library and normalises to NFKD before removal.

    Args:
        input_text: The text you want to remove diacritics from.
        errors (optional): Default = 'raise' which raises all errors. 'ignore' ignores errors and returns original value, 'coerce' returns None.

    Raises:
        TypeError: Raised if input_text is not a string.

    Returns:
        Text with accents removed e.g. 'Chloë' -> 'Chloe'.
    """
    try:
        if not isinstance(input_text, str):
            raise TypeError(f"Input must be a string, not {type(input_text).__name__}")
        nfkd_form = unicodedata.normalize("NFKD", input_text)
        return "".join([c for c in nfkd_form if unicodedata.category(c) != "Mn"])
    except:
        if errors == "coerce":
            return None
        if errors == "ignore":
            return input_text
        raise


def remove_punctuation(
    text: str, punctuation: str = PUNCTUATION, errors: str = "raise"
) -> str | None:
    r"""Removes all punctuation except for hyphens and apostrophes from text. Useful for cleaning names.

    Args:
        text (str): Text you wish to remove punctuation from.
        punctuation (optional): String containing all punctuation except for hyphens and apostrophes. Can be overridden with your own version if you want to exclude other types of punctuation. Should be one string of all chars to remove. Default includes the following chars: !@#£$%^&*()_=+`~,.<>/?;:"\|[]
        errors (optional): Default = 'raise' which raises all errors. 'ignore' ignores errors and returns original value, 'coerce' returns None.

    Raises:
        TypeError: Raised if text is not a string.

    Returns:
        Text with all punctuation except hyphens and apostrophes removed e.g. 'Jane! Doe.' -> 'Jane Doe'
    """
    try:
        if not isinstance(text, str):
            raise TypeError(f"Input must be a string, not {type(text).__name__}")
        table = str.maketrans(punctuation, " " * len(punctuation))
        text = text.strip().translate(table)
        cleaned = re.sub(r"\s+", " ", text).strip()
        return cleaned
    except:
        if errors == "coerce":
            return None
        if errors == "ignore":
            return text
        raise



import os
import pandas as pd

from .exceptions import ColumnsNotUnique
from .core import _to_snake


# Utility Functions
def convert_col_snake_case(df: pd.DataFrame) -> pd.DataFrame:
    """_summary_

    Args:
        df (pd.DataFrame): _description_

    Raises:
        TypeError: _description_
        ColumnsNotUnique: _description_
        ColumnsNotUnique: _description_

    Returns:
        pd.DataFrame: _description_
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError(f"df must be a DataFrame, not {type(df).__name__}")

    columns = list(df.columns)

    if len(set(columns)) != len(columns):
        duplicates = set([x for x in columns if columns.count(x) > 1])
        raise ColumnsNotUnique(f"DataFrame has duplicate columns: {duplicates}")

    new_df = df.copy()
    new_cols = [_to_snake(c) for c in new_df.columns]
    if len(set(new_cols)) != len(new_cols):
        # Identify which names are duplicates for a better error message
        duplicates = set([x for x in new_cols if new_cols.count(x) > 1])
        raise ColumnsNotUnique(
            f"Snake-case conversion created duplicate columns: {duplicates}"
        )

    new_df.columns = new_cols

    return new_df


# File Manipulation
def get_excel_filepaths_in_folder(
    input_dir: str, print_to_terminal: bool = False
) -> list[str]:
    """Returns a list of filepaths to Excel files (with the extension .xlsx or .xls) in a given folder.

    Args:
        input_dir: The directory you want to get the filepaths from.
        print_to_terminal (optional): Defaults to False. Set to True if you want the terminal to print messages about the file processing.

    Raises:
        FileNotFoundError: Raises errors if the directory does not exist.

    Returns:
        List: A list of filepaths from the specified folder. Returns empty list if there are no Excel files in the folder.
    """

    # Helper function to control if messages are printed to terminal
    def log(message: str):
        if print_to_terminal:
            print(message)

    # Checks directory exists before function starts
    if not os.path.exists(input_dir):
        raise FileNotFoundError(f"The directory '{input_dir}' does not exist.")

    log(f"\nDiscovering files in '{input_dir}'")
    excel_files = 0
    filepaths = []  # empty list the filepaths will be added to
    # Iterate over all files in the given folder
    for filename in os.listdir(input_dir):  # for each file in the folder
        file_path = os.path.join(input_dir, filename)  # create the filepath
        # Check if it's an actual file and an Excel file
        if os.path.isfile(file_path) and filename.lower().endswith((".xlsx", ".xls")):
            excel_files += 1
            filepaths.append(file_path)  # add each filepath to the filepaths list
            log(f"  Found Excel file: {filename} and added to processing list.")
        else:
            log(f"  Skipping non-Excel file: {filename}")
    if not filepaths:
        log("No Excel files found to process.")
        return []
    return filepaths



# Import internal libraries
import re

# Import helper functions
from heat_helper.core import _is_valid_postcode
from heat_helper.exceptions import InvalidPostcodeError


def format_postcode(postcode: str, errors: str = "raise") -> str | None:
    """Attempts to clean a postcode to conform to UK standard.

    Args:
        postcode: Text you want to clean.
        errors: default = 'raise' which raises all errors. 'ignore' returns orignal value, 'coerce' attempts to turn postcode into string to run the function, if can't be run returns None.

    Raises:
        TypeError: Raised if postcode is not a string.
        InvalidPostcodeError: Raised if postcode is not a valid length (5 to 7 chars).

    Returns:
        Cleaned postcode.
    """
    # checks is postcode is a string and returns original if not
    try:
        if not isinstance(postcode, str):
            if errors == "coerce":
                postcode = str(postcode)
            else:
                raise TypeError(
                    f"Postcode must be a string, not {type(postcode).__name__}"
                )

        clean = re.sub(r"\s+", "", postcode)
        clean = clean.upper().strip()
        formatted = f"{clean[:-3]} {clean[-3:]}"

        if not _is_valid_postcode(formatted):
            raise InvalidPostcodeError(postcode)

        return formatted
    except (InvalidPostcodeError, TypeError):
        if errors == "ignore":
            return postcode
        if errors == "coerce":
            return None
        raise
